### This script creates the Fitz-Hugh Nagumo model. It sets up the 
	### odes and begins preparation for the MCMC to be set-up/run


### Load the "deSolve" package

library(deSolve)


### Creating the system of ordinary differential equations for the
	### Fitz-Hugh Nagumo model

### Species: There are 2 species

### Species in order are:
### 1. V (Voltage)
### 2. R (Recovery Variable)


### Parameters: There are 3 parameters
### 1. Alpha
### 2. Beta
### 3. Gamma


### Calculating gradients for the Fitz-Hugh Nagumo model. This consists of 
	### setting up the system of odes. dydt contains the values of dy by dt
	### (differentiating y with respect to t, where y is the response i.e.
	### the species, and t is time). Species2 stands for the variable X, 
	### and params is a vector of the parameter values

gradientsFitzHughNagumoModel <- function(params,Species2,timepoints=NULL,
							auxVars,species)
{

	dydt <- switch(species,

		params[3]*(Species2[,1] + Species2[,2] - ((Species2[,1]^3)/3))
	,
		- (Species2[,1] - params[1] + params[2]*Species2[,2])/params[3]

			  		)

	return(dydt)
}


### Creating function odeFitzHughNagumo. This will be called in a later stage,
	### during the population MCMC stage and is similar to the gradient
	### specification above. If using the function "c", then each element
	### will be one equation. Species here stands for the variable X (or x)

odeFitzHughNagumoModel <- function(t,Species,params)
{
	dydt <- c(

		params[3]*(Species[1] + Species[2] - ((Species[1]^3)/3))
	,
		- (Species[1] - params[1] + params[2]*Species[2])/params[3]

			      )

	grads = list()
	grads[[1]] = dydt

  	return(grads)
}


### Creation of simulation Fitz-Hugh Nagumo. It uses the function "ode" in the
	### cran package "deSolve" to solve the odeFitzHughNagumo model function
	### above. xInit is the variable containing the initial values for the
	### species concentrations, t is the timepoints at which each species is
	### simulated and params is the vector of parameter values

simulateFitzHughNagumoModel <- function(xInit,t,params,...)
{
	solution <- ode(xInit,t,odeFitzHughNagumoModel,params,...)

	return(solution)
}










