### This file sets up and runs the MCMC.

### Note: To change the prior, search for FLAG PRIOR in the file
	### population_mcmcTemperingMismatch.R


### The following packages must be installed: "gptk", "gdata", "deSolve"

for (DATA in 1:10)	# This creates a loop to run over the different datasets
				# the set-up and implementation of the MCMC
{

### Choose the average signal to noise ratio that generated the data. Note
	### that this is only used to load the correct dataset
averageSNR <- 10
#averageSNR <- 5

noiseInferRun <- FALSE # Specify whether the error is inferred or held at a
					# fixed value


dataset.i <- DATA # Dataset number - used for loading the correct dataset,
				# since it is usual to simulate a number of datasets
				# in order to observe the average performance

temperatureExponentChains <- 5 # This is the value of the power to set the 
						# various power posterior distributions
						# for the parallel tempering, see equation
						# 18 in the paper "Approximate parameter 
						# inference in systems biology using 
						# gradient matching: a comparative
						# evaluation", Biomedical Engineering
						# Online, for details. The power 5 is
						# used as recommended in the paper
						# "Marginal likelihood estimation via
						# power posteriors." by Friel and Pettit

### Translate the average signal to noise ratio to the standard deviation of
	### the error
if(averageSNR == 10)
{
	noise <- 0.38
}
if(averageSNR ==5)
{
	noise <- 0.535
}




source("MCMC Source/population_mcmcTemperingMismatch.R") # Specify the
										# location for the
										# MCMC files

source("Fitz-Hugh Nagumo Model.R")  # Specify the location of the ODE model



### Load the dataset that one wants to examine. 

load(paste("FhN SD Noise ",noise," Average SNR ",averageSNR," ",dataset.i,".R",sep=""))


### For ease of coding, create variable params, from the dataset params

params <- dataset$params



submodels <- 1:2
	# There will be as many submodels as species, since each species is a
		# submodel


submodel.params <- list(


		c(3)
	,
		c(1,2,3)
)

		# Each element of the list "submodel.params", contains the numbers
			# of which parameters make up the species of element i. For
			# instance, element 2 corresponds to species 2 and the
			# parameters that make it up are alpha, beta, gamma which
			# are parameters 1, 2, 3 in this case



t <- seq(0,10,length.out=25) # Specify the timepoints for the observations

time <- as.matrix(t)
	# Matrix version of the timepoints in case useful




### Since we assume a zero mean GP, the data needs to have the means
	### subtracted. These need to be stored and passed to the MCMC, since
	### the ODEs need to use the data WITH the original means, in order to
	### make correct predictions

dataConstants <- matrix(,nrow=length(t),ncol=max(submodels))
for (i in 1:ncol(dataConstants))
{
	dataConstants[,i] <- mean(dataset$data[,i])
}


dataset$data <- dataset$data - dataConstants

FitzHughNagumoData <- dataset$data

data.true <- dataset$data.true # Used for plotting



### Set up the auxilarly variables. These are called within the population
	### MCMC function. Defaults are as they have been used with previous
	### work, but can be altered


auxVars <- list(y.true=data.true,modelName="FitzHughNagumoModel",speciesList=submodels,
			covtype="rbf",observedSpeciesList=submodels,
			constant=dataConstants,sigmaInfer=noiseInferRun,
			sigmaTrue=noise)


### Specify the parameters to be inferred. 


inferredParams <- 1:range(submodel.params)[2]



### Specifiy how the initial parameters are chosen. Sometimes it is desired to
	### set values of the initial parameters to the true parameters used, so
	### that the MCMC algorithm has a good place to start and sometimes is
	### desired to randomly draw the starting values. This means it is more
	### difficult for MCMC algorithm to learn the parameters, and previous
	### studies can indicate values that may prove to be a good stating
	### point, and thus the draw can be constrained to similar values. The
	### default will be for the initial parameters to be drawn from the
	### prior

paramsInit <- c()

for (i in 1:range(submodel.params)[2])
{
	if (i < 3)
	{
		paramsInit[i] <- rnorm(1,0,0.4)
	}
	if (i == 3)
	{
		paramsInit[i] <- rchisq(1,2)
	}
}


paramNames <- c("Alpha","Beta","Gamma")


names(paramsInit) <- paramNames

### Specify the tuning that will take place in the MCMC algorithm

proposalTuning <- matrix(0.01,nrow=10,ncol=range(submodel.params)[2])
	# A matrix of tuning values, with the matrix having the same number of
		# columns as the number of parameters

proposalXTuning <- matrix(0.01,nrow=10,ncol=dim(data.true)[2])
	# A matrix of X tuning values, with the matrix having the same number
		# of columns as the number of species

proposalNoiseTuning <- matrix(0.1,nrow=10,ncol=dim(data.true)[2])
	# A matrix of noise tuning values, with the matrix having the same
		# number of columns as the number of species

proposalGPTuning <- matrix(0.05,nrow=3,ncol=dim(data.true)[2])
	# A matrix of GP tuning values, with the matrix having the same number
		# of columns as the number of species

chainNum <- 10
	# Number of parallel chains/temperatures


### Specify the gradient mismatch parameter, see equation 12 in the paper
	### "Approximate parameter inference in systems biology using 
	### gradient matching: a comparative evaluation", Biomedical
	### Engineering Online, for details. NOTE: The parameter is called gamma
	### in the paper, but here it is called lambda. The code was written
	### well before the paper and this is the cause for the differences in
	### the names

Base <- 10
	# The base that will be used to create the increments for the log scale

startLambda <- 1
	# The starting value for lambda (gamma), corresponding to the lowest
		# chain/temperature

LambdaTemperatures <- c()
LambdaTemperatures[1] <- startLambda

for (i in 1:(chainNum-1))
{
	LambdaTemperatures[i+1] <- LambdaTemperatures[1]*(Base^i)
}

Lambda <- matrix(LambdaTemperatures,nrow=chainNum,ncol=dim(data.true)[2])
	# Lambda is now a matrix of the different Lambda (gamma) temperatures,
		# with number of rows corresponding to number of
		# chains/temperatures and number of columns corresponding to the
		# number of species

Lambda <- 1/Lambda


### Specify the save files, for the results of the population MCMC to be
	### stored


if(noiseInferRun==FALSE){
saveFile <- paste("LB10 FhN True Noise Fixed RBF Noise ",noise," Average SNR ",averageSNR," Dataset ",dataset.i,".R",sep="")}

if(noiseInferRun==TRUE){
saveFile <- paste("LB10 FhN Noise Inferred RBF Noise ",noise," Average SNR ",averageSNR," Dataset ",dataset.i,".R",sep="")}




### Specify the options for the population MCMC algorithm to operate. The
	### below are the defaults

options <- list(inferredParams=inferredParams,samplingStrategy="MCMC",
			samplingStrategyX="mixed",iterations=300000,
			proposalTuning=proposalTuning,inferX=TRUE,
	proposalXTuning=proposalXTuning,proposalNoiseTuning=proposalNoiseTuning,
			proposalGPTuning=proposalGPTuning,burnin=1,
			chainNum=chainNum,paramsInit=paramsInit,
			noiseBurnin=1,showPlot=T,
			saveFile=saveFile,lambda.max=0.1,
			showProgress=F,keepInit=F,
			explicit=F,allowNeg=T,
			gpInit=T,lambda=Lambda,temps=temperatureExponentChains
)

### Run population MCMC algorithm to infer the parameters of interest. At
	### present, only the top chain is saved (the posterior) and every
	### 10,000 iterations, the function saves the results of the MCMC to the
	### location specified in "saveFile". The log likelihood is recorded
	### for all chains and saved as "lLAllChains". The MCMC performs some
	### thinning by only recording every 100 samples. The inferred ODE
	### parameters are saved as "paramsRec". 

paramsMCMC <- doMCMC(time,FitzHughNagumoData,auxVars,options)

### Save the final results of the MCMC to the location specified in "saveFile"

save(paramsMCMC,file=saveFile)

rm(list=ls()) # Remove variables from workspace to avoid any crossover from
			# different datasets
}



